from .equityTracking_client import EquityTrackingClient
from .drawdownListener import DrawdownListener
import pytest
from mock import MagicMock, AsyncMock
domain_client = MagicMock()
equity_tracking_client = EquityTrackingClient(domain_client)
token = 'header.payload.sign'


class Listener(DrawdownListener):
    async def on_drawdown(self, drawdown_event):
        pass


listener: DrawdownListener = None


@pytest.fixture(autouse=True)
async def run_around_tests():
    global domain_client
    domain_client = MagicMock()
    domain_client.request_api = AsyncMock()
    domain_client.token = token
    global equity_tracking_client
    equity_tracking_client = EquityTrackingClient(domain_client)
    global listener
    listener = Listener()


class TestEquityTrackingClient:
    @pytest.mark.asyncio
    async def test_create_drawdown_tracker(self):
        """Should create drawdown tracker."""
        expected = {'id': 'trackerId'}
        tracker = {'name': 'trackerName'}
        domain_client.request_api = AsyncMock(return_value=expected)
        actual = await equity_tracking_client.create_drawdown_tracker('accountId', tracker)
        assert actual == expected
        domain_client.request_api.assert_called_with({
            'url': '/users/current/accounts/accountId/drawdown-trackers',
            'method': 'POST',
            'body': tracker
        })

    @pytest.mark.asyncio
    async def test_retrieve_drawdown_trackers(self):
        """Should retrieve drawdown trackers."""
        expected = [{'name': 'trackerName'}]
        domain_client.request_api = AsyncMock(return_value=expected)
        actual = await equity_tracking_client.get_drawdown_trackers('accountId')
        assert actual == expected
        domain_client.request_api.assert_called_with({
            'url': '/users/current/accounts/accountId/drawdown-trackers',
            'method': 'GET'
        })

    @pytest.mark.asyncio
    async def test_update_drawdown_tracker(self):
        """Should update drawdown tracker."""
        update = {'name': 'newTrackerName'}
        await equity_tracking_client.update_drawdown_tracker('accountId', 'trackerId', update)
        domain_client.request_api.assert_called_with({
            'url': '/users/current/accounts/accountId/drawdown-trackers/trackerId',
            'method': 'PUT',
            'body': update
        })

    @pytest.mark.asyncio
    async def test_delete_drawdown_tracker(self):
        """Should delete drawdown tracker."""
        await equity_tracking_client.delete_drawdown_tracker('accountId', 'trackerId')
        domain_client.request_api.assert_called_with({
            'url': '/users/current/accounts/accountId/drawdown-trackers/trackerId',
            'method': 'DELETE'
        })

    @pytest.mark.asyncio
    async def test_retrieve_drawdown_events(self):
        """Should retrieve drawdown events."""
        expected = [{
            'sequenceNumber': 1,
            'accountId': 'accountId',
            'trackerId': 'trackerId',
            'period': 'day',
            'startBrokerTime': '2022-04-08 00:00:00.000',
            'endBrokerTime': '2022-04-08 23:59:59.999',
            'brokerTime': '2022-04-08 09:36:00.000',
            'absoluteDrawdown': 250,
            'relativeDrawdown': 0.25
        }]
        domain_client.request_api = AsyncMock(return_value=expected)

        actual = await equity_tracking_client.get_drawdown_events(
            '2022-04-08 09:36:00.000', '2022-04-08 10:36:00.000', 'accountId', 'trackerId', 100)
        assert actual == expected
        domain_client.request_api.assert_called_with({
            'url': '/users/current/drawdown-events/by-broker-time',
            'params': {
                'startBrokerTime': '2022-04-08 09:36:00.000',
                'endBrokerTime': '2022-04-08 10:36:00.000',
                'accountId': 'accountId',
                'trackerId': 'trackerId',
                'limit': 100
            },
            'method': 'GET'
        })

    @pytest.mark.asyncio
    async def test_retrieve_drawdown_statistics(self):
        """Should retrieve drawdown statistics."""
        expected = [{
            'period': 'day',
            'startBrokerTime': '2022-04-08 00:00:00.000',
            'endBrokerTime': '2022-04-08 23:59:59.999',
            'initialBalance': 1000,
            'maxDrawdownTime': '2022-04-08 09:36:00.000',
            'maxAbsoluteDrawdown': 250,
            'maxRelativeDrawdown': 0.25,
            'thresholdExceeded': True
        }]
        domain_client.request_api = AsyncMock(return_value=expected)

        actual = await equity_tracking_client.get_drawdown_statistics('accountId', 'trackerId',
                                                                      '2022-04-08 09:36:00.000', 100)
        assert actual == expected
        domain_client.request_api.assert_called_with({
            'url': '/users/current/accounts/accountId/drawdown-trackers/trackerId/statistics',
            'params': {'startTime': '2022-04-08 09:36:00.000', 'limit': 100},
            'method': 'GET'
        })

    @pytest.mark.asyncio
    async def test_retrieve_equity_chart(self):
        """Should retrieve equity chart."""
        expected = [{
            'startBrokerTime': '2022-04-08 00:00:00.000',
            'endBrokerTime': '2022-04-08 23:59:59.999',
            'averageBalance': 1050,
            'minBalance': 100,
            'maxBalance': 2000,
            'averageEquity': 1075,
            'minEquity': 50,
            'maxEquity': 2100
        }]
        domain_client.request_api = AsyncMock(return_value=expected)

        actual = await equity_tracking_client.get_equity_chart('accountId', '2022-04-08 09:36:00.000',
                                                               '2022-04-08 10:36:00.000')
        assert actual == expected
        domain_client.request_api.assert_called_with({
            'url': '/users/current/accounts/accountId/equity-chart',
            'params': {
                'startTime': '2022-04-08 09:36:00.000',
                'endTime': '2022-04-08 10:36:00.000'
            },
            'method': 'GET'
        })

    @pytest.mark.asyncio
    async def test_add_drawdown_listener(self):
        """Should add drawdown listener."""
        call_stub = MagicMock()
        equity_tracking_client._drawdownListenerManager.add_drawdown_listener = call_stub
        equity_tracking_client.add_drawdown_listener(listener, 'accountId', 'trackerId', 1)
        call_stub.assert_called_with(listener, 'accountId', 'trackerId', 1)

    @pytest.mark.asyncio
    async def test_remove_drawdown_listener(self):
        """Should remove drawdown listener."""
        call_stub = MagicMock()
        equity_tracking_client._drawdownListenerManager.remove_drawdown_listener = call_stub
        equity_tracking_client.remove_drawdown_listener('id')
        call_stub.assert_called_with('id')



from .clients.httpClient import HttpClient
from .clients.domain_client import DomainClient
from .clients.equityTracking.equityTracking_client import EquityTrackingClient
from .models import format_error
from typing_extensions import TypedDict
from typing import Optional


class RetryOpts(TypedDict):
    retries: Optional[int]
    """Maximum amount of request retries, default value is 5."""
    minDelayInSeconds: Optional[float]
    """Minimum delay in seconds until request retry, default value is 1."""
    maxDelayInSeconds: Optional[float]
    """Maximum delay in seconds until request retry, default value is 30."""


class RiskManagementOptions(TypedDict):
    """Risk management SDK options."""
    domain: Optional[str]
    """Domain to connect to."""
    extendedTimeout: Optional[float]
    """Timeout for extended http requests in seconds."""
    requestTimeout: Optional[float]
    """Timeout for http requests in seconds."""
    retryOpts: Optional[RetryOpts]
    """Options for request retries."""


class RiskManagement:
    """MetaApi risk management API SDK."""

    def __init__(self, token: str, opts: RiskManagementOptions = None):
        """Inits class instance.

        Args:
            token: Authorization token.
            opts: Connection options.
        """
        opts: RiskManagementOptions = opts or {}
        domain = opts['domain'] if 'domain' in opts else 'agiliumtrade.agiliumtrade.ai'
        request_timeout = opts['requestTimeout'] if 'requestTimeout' in opts else 10
        request_extended_timeout = opts['extendedTimeout'] if 'extendedTimeout' in opts else 70
        retry_opts = opts['retryOpts'] if 'retryOpts' in opts else {}
        http_client = HttpClient(request_timeout, request_extended_timeout, retry_opts)
        self._domainClient = DomainClient(http_client, token, 'risk-management-api-v1', domain)
        self._equityTrackingClient = EquityTrackingClient(self._domainClient)

    @property
    def risk_management_api(self) -> EquityTrackingClient:
        """Returns CopyFactory configuration API.

        Returns:
            Configuration API.
        """
        return self._equityTrackingClient

    @staticmethod
    def format_error(err: Exception):
        """Formats and outputs metaApi errors with additional information.

        Args:
            err: Exception to process.
        """
        return format_error(err)



from datetime import datetime
from copy import copy


class DomainClient:
    """Connection URL and request managing client"""

    def __init__(self, http_client, token: str, api_path: str, domain: str = None):
        """Inits domain client instance.

        Args:
            http_client: HTTP client.
            token: Authorization token.
            api_path: Api URL part.
            domain: Domain to connect to, default is agiliumtrade.agiliumtrade.ai.
        """
        self._httpClient = http_client
        self._apiPath = api_path
        self._domain = domain or 'agiliumtrade.agiliumtrade.ai'
        self._token = token
        self._urlCache = None
        self._regionCache = []
        self._regionIndex = 0

    @property
    def domain(self) -> str:
        """Returns domain client domain.

        Returns:
            Client domain.
        """
        return self._domain

    @property
    def token(self) -> str:
        """Returns domain client token.

        Returns:
            Client token.
        """
        return self._token

    async def request_api(self, opts: dict, is_extended_timeout: bool = False):
        """Sends an authorized json API request.

        Args:
            opts: Options request options.
            is_extended_timeout: Whether to run the request with an extended timeout.

        Returns:
            Request result.
        """
        await self._update_host()
        try:
            request_opts = copy(opts)
            request_opts['headers'] = opts['headers'] if 'headers' in opts else {'auth-token': self._token}
            request_opts['url'] = self._urlCache['url'] + request_opts['url']
            return await self._httpClient.request(request_opts, is_extended_timeout)
        except Exception as err:
            if err.__class__.__name__ not in ['ConflictException', 'InternalException', 'ApiException',
                                              'ConnectTimeout']:
                raise err
            else:
                if len(self._regionCache) == self._regionIndex + 1:
                    raise err
                else:
                    self._regionIndex += 1
                    return await self.request_api(opts, is_extended_timeout)

    async def request(self, opts: dict):
        """Sends an http request.

        Args:
            opts: Request options.

        Returns:
            Request result.
        """
        return await self._httpClient.request(opts)

    async def _update_host(self):
        if not self._urlCache or self._urlCache['lastUpdated'] < datetime.now().timestamp() - 60 * 10:
            await self._update_regions()
            url_settings = await self._httpClient.request({
                'url': f'https://mt-provisioning-api-v1.{self._domain}/users/current/servers/mt-client-api',
                'method': 'GET',
                'headers': {
                    'auth-token': self._token
                }
            })
            self._urlCache = {
                'url': f'https://{self._apiPath}.{self._regionCache[self._regionIndex]}.{url_settings["domain"]}',
                'domain': url_settings['domain'],
                'lastUpdated': datetime.now().timestamp()
            }
        else:
            self._urlCache = {
                'url': f'https://{self._apiPath}.{self._regionCache[self._regionIndex]}.{self._urlCache["domain"]}',
                'domain': self._urlCache['domain'],
                'lastUpdated': datetime.now().timestamp()
            }

    async def _update_regions(self):
        self._regionIndex = 0
        self._regionCache = await self._httpClient.request({
            'url': f'https://mt-provisioning-api-v1.{self._domain}/users/current/regions',
            'method': 'GET',
            'headers': {
                'auth-token': self._token
            },
        })



from .httpClient import HttpClient
import re
import pytest
import respx
from datetime import datetime
import json
from httpx import Response
from ..models import format_date
httpClient: HttpClient = None
test_url = 'http://example.com'
opts = {}


@pytest.fixture(autouse=True)
async def run_around_tests():
    global httpClient
    httpClient = HttpClient()
    global opts
    opts = {'url': test_url}
    yield


class TestHttpClient:
    @pytest.mark.asyncio
    async def test_load(self):
        """Should load HTML page from example.com"""
        response = await httpClient.request(opts)
        text = response.text
        assert re.search('doctype html', text)

    @pytest.mark.asyncio
    async def test_load_failover(self):
        """Should load HTML page from example.com with failover"""
        response = await httpClient.request_with_failover(opts)
        text = response.text
        assert re.search('doctype html', text)

    @pytest.mark.asyncio
    async def test_not_found(self):
        """Should return NotFound exception if server returns 404"""
        opts = {
            'url': f'{test_url}/not-found'
        }
        try:
            await httpClient.request_with_failover(opts)
            raise Exception('NotFoundException is expected')
        except Exception as err:
            assert err.__class__.__name__ == 'NotFoundException'

    @pytest.mark.asyncio
    async def test_timeout(self):
        """Should return ConnectTimeout exception if request is timed out"""
        httpClient = HttpClient(0.001, 60, {'retries': 2, 'minDelayInSeconds': 0.05, 'maxDelayInSeconds': 0.2})
        try:
            await httpClient.request_with_failover(opts)
            raise Exception('ConnectTimeout is expected')
        except Exception as err:
            assert err.__class__.__name__ == 'ConnectTimeout'

    @respx.mock
    @pytest.mark.asyncio
    async def test_retry_on_api_exception(self):
        """Should retry request on fail with ApiException exception."""
        respx.get(test_url).mock(side_effect=[Response(502), Response(502),
                                              Response(200, content=json.dumps('response'))])
        response = await httpClient.request_with_failover(opts)
        assert response == 'response'
        assert respx.get(test_url).call_count == 3

    @respx.mock
    @pytest.mark.asyncio
    async def test_retry_on_internal_exception(self):
        """Should retry request on fail with InternalException exception."""
        respx.get(test_url).mock(side_effect=[Response(500), Response(500),
                                              Response(200, content=json.dumps('response'))])
        response = await httpClient.request_with_failover(opts)
        assert response == 'response'
        assert respx.get(test_url).call_count == 3

    @respx.mock
    @pytest.mark.asyncio
    async def test_return_error_on_retry_limit_exceeded(self):
        """Should return error if retry limit exceeded."""
        respx.get(test_url).mock(side_effect=Response(502))
        httpClient = HttpClient(10, 60, {'retries': 2})
        try:
            await httpClient.request_with_failover(opts)
            raise Exception('ApiException is expected')
        except Exception as err:
            assert err.__class__.__name__ == 'ApiException'
        assert respx.get(test_url).call_count == 3

    @respx.mock
    @pytest.mark.asyncio
    async def test_not_retry_if_exception_not_internal_exception_or_api_exception(self):
        """Should not retry if exception is neither InternalException nor ApiException."""
        error = {
            'id': 1,
            'error': 'error',
            'message': 'test',
            'details': [{'parameter': 'password', 'value': 'wrong', 'message': 'Invalid value'}]
        }
        respx.get(test_url).mock(side_effect=[Response(400, json=error), Response(400, json=error), Response(204)])
        try:
            await httpClient.request_with_failover(opts)
            raise Exception('ValidationException is expected')
        except Exception as err:
            assert err.__class__.__name__ == 'ValidationException'
            assert err.__str__() == 'test, check error.details for more information'
            assert err.details == error['details']
        assert respx.get(test_url).call_count == 1

    def get_too_many_requests_error(self, sec):
        date = datetime.now().timestamp()
        date += sec
        recommended_retry_time = format_date(datetime.fromtimestamp(date))
        return Response(429, content=json.dumps({"metadata": {"recommendedRetryTime": recommended_retry_time}}))

    @respx.mock
    @pytest.mark.asyncio
    async def test_retry_after_wait_on_too_many_requests_error(self):
        """Should retry request after waiting on fail with TooManyRequestsException error."""
        respx.get(test_url).mock(side_effect=[self.get_too_many_requests_error(2),
                                              self.get_too_many_requests_error(3),
                                              Response(200, content=json.dumps('response'))])
        response = await httpClient.request_with_failover(opts)
        assert response == 'response'
        assert respx.get(test_url).call_count == 3

    @respx.mock
    @pytest.mark.asyncio
    async def test_return_error_if_recommended_time_too_long(self):
        """Should return error if recommended retry time is too long."""
        respx.get(test_url).mock(side_effect=[self.get_too_many_requests_error(2),
                                              self.get_too_many_requests_error(300),
                                              Response(200, content=json.dumps('response'))])
        try:
            await httpClient.request_with_failover(opts)
            raise Exception('TooManyRequestsException is expected')
        except Exception as err:
            assert err.__class__.__name__ == 'TooManyRequestsException'
        assert respx.get(test_url).call_count == 2

    @respx.mock
    @pytest.mark.asyncio
    async def test_not_count_retrying_too_many_requests_exception(self):
        """Should not count retrying TooManyRequestsException error."""
        respx.get(test_url).mock(side_effect=[self.get_too_many_requests_error(2), Response(502),
                                              Response(200, content=json.dumps('response'))])
        httpClient = HttpClient(10, 60, {'retries': 1})
        response = await httpClient.request_with_failover(opts)
        assert response == 'response'
        assert respx.get(test_url).call_count == 3

    @respx.mock
    @pytest.mark.asyncio
    async def test_wait_for_retry_after_header(self):
        """Should wait for the retry-after header time before retrying."""
        httpClient = HttpClient()
        respx.get(test_url).mock(side_effect=[
            Response(202, headers={'retry-after': '3'}),
            Response(202, headers={'retry-after': '3'}),
            Response(200, content=json.dumps('response'))])
        response = await httpClient.request_with_failover(opts)
        assert response == 'response'
        assert respx.get(test_url).call_count == 3

    @respx.mock
    @pytest.mark.asyncio
    async def test_return_error_if_retry_after_time_too_long(self):
        """Should return TimeoutException error if retry-after header time is too long."""
        httpClient = HttpClient(10, 60, {'maxDelayInSeconds': 3})
        respx.get(test_url).mock(side_effect=[
            Response(202, headers={'retry-after': '30'}),
            Response(202, headers={'retry-after': '30'}),
            Response(200, content=json.dumps('response'))])
        try:
            await httpClient.request_with_failover(opts)
            raise Exception('TimeoutException is expected')
        except Exception as err:
            assert err.__class__.__name__ == 'TimeoutException'
            assert err.args[0] == 'Timed out waiting for the response'
        assert respx.get(test_url).call_count == 1

    @respx.mock
    @pytest.mark.asyncio
    async def test_return_error_if_timed_out_to_retry(self):
        """Should return TimeoutException error if timed out to retry."""
        respx.get(test_url).mock(side_effect=Response(202, headers={'retry-after': '1'}))
        httpClient = HttpClient(10, 60, {'maxDelayInSeconds': 2, 'retries': 3})
        try:
            await httpClient.request_with_failover(opts)
            raise Exception('TimeoutException is expected')
        except Exception as err:
            assert err.__class__.__name__ == 'TimeoutException'
            assert err.args[0] == 'Timed out waiting for the response'
        assert respx.get(test_url).call_count == 6



from .httpClient import HttpClient
from .domain_client import DomainClient
from mock import AsyncMock, ANY
import pytest
from freezegun import freeze_time
import respx
import json
from httpx import Response
from datetime import datetime

http_client = HttpClient()
token = 'header.payload.sign'
domain_client = DomainClient(http_client, token, 'risk-management-api-v1')
request_mock = AsyncMock()
expected = [{'_id': 'ABCD'}]
start_time = '2020-10-05 10:00:00.000'

request_call: respx.Route
regions_call: respx.Route
host_call: respx.Route
request_url = 'https://risk-management-api-v1.vint-hill.agiliumtrade.agiliumtrade.ai/some/rest/api'
regions_url = 'https://mt-provisioning-api-v1.agiliumtrade.agiliumtrade.ai/users/current/regions'
host_url = 'https://mt-provisioning-api-v1.agiliumtrade.agiliumtrade.ai/users/current/servers/mt-client-api'
opts = {
    'url': '/some/rest/api',
    'method': 'GET'
}
host = {
    'host': 'https://copyfactory-api-v1',
    'lastUpdated': datetime.now().timestamp(),
    'regions': ['vint-hill'],
    'domain': 'agiliumtrade.ai'
}


@pytest.fixture(autouse=True)
async def run_around_tests():
    global http_client
    http_client = HttpClient()
    global domain_client
    domain_client = DomainClient(http_client, token, 'risk-management-api-v1')

    global request_call
    request_call = respx.get(request_url).mock(return_value=Response(200, content=json.dumps(expected)))

    global regions_call
    regions_call = respx.get(regions_url)\
        .mock(return_value=Response(200, content=json.dumps(['vint-hill', 'us-west'])))

    global host_call
    host_call = respx.get(host_url)\
        .mock(return_value=Response(200, content=json.dumps({'domain': 'agiliumtrade.agiliumtrade.ai'})))

    global host
    host = {
        'host': 'https://copyfactory-api-v1',
        'lastUpdated': datetime.now().timestamp(),
        'regions': ['vint-hill'],
        'domain': 'agiliumtrade.ai'
    }
    yield


class TestDomainClient:

    @respx.mock
    @pytest.mark.asyncio
    async def test_execute_request(self):
        """Should execute request."""
        response = await domain_client.request_api(opts)
        assert response == expected

    @respx.mock
    @pytest.mark.asyncio
    async def test_use_cached_url(self):
        """Should use cached url on repeated request."""
        await domain_client.request_api(opts)
        response = await domain_client.request_api(opts)
        assert response == expected
        assert host_call.call_count == 1
        assert regions_call.call_count == 1

    @respx.mock
    @pytest.mark.asyncio
    async def test_request_url_again_if_expired(self):
        """Should request url again if expired."""
        with freeze_time(start_time) as frozen_datetime:
            await domain_client.request_api(opts)
            frozen_datetime.tick(610)
            response = await domain_client.request_api(opts)
            assert response == expected
            assert host_call.call_count == 2
            assert regions_call.call_count == 2

    @respx.mock
    @pytest.mark.asyncio
    async def test_return_request_error(self):
        """Should return request error."""
        request_call.mock(return_value=Response(400))
        try:
            await domain_client.request_api(opts)
            pytest.fail()
        except Exception as err:
            assert err.__class__.__name__ == 'ValidationException'

    @respx.mock
    @pytest.mark.asyncio
    async def test_return_error_if_failed_host(self):
        """Should return error if failed to get host."""
        host_call.mock(return_value=Response(400))
        try:
            await domain_client.request_api(opts)
            pytest.fail()
        except Exception as err:
            assert err.__class__.__name__ == 'ValidationException'

    @respx.mock
    @pytest.mark.asyncio
    async def test_return_error_if_failed_regions(self):
        """Should return error if failed to get regions."""
        regions_call.mock(return_value=Response(400))
        try:
            await domain_client.request_api(opts)
            pytest.fail()
        except Exception as err:
            assert err.__class__.__name__ == 'ValidationException'

    @respx.mock
    @pytest.mark.asyncio
    async def test_try_another_region_if_first_failed(self):
        """Should try another region if the first failed."""
        request_call.mock(return_value=Response(500))
        us_west_call = respx.get('https://risk-management-api-v1.us-west.agiliumtrade.agiliumtrade.ai/some/rest/api')\
            .mock(return_value=Response(200, content=json.dumps(expected)))
        response = await domain_client.request_api(opts)
        assert response == expected
        assert us_west_call.call_count == 1
        assert host_call.call_count == 1
        assert regions_call.call_count == 1

    @respx.mock
    @pytest.mark.asyncio
    async def test_return_error_if_all_regions_failed(self):
        """Should return error if all regions failed."""
        request_call.mock(return_value=Response(500))
        respx.get('https://risk-management-api-v1.us-west.agiliumtrade.agiliumtrade.ai/some/rest/api') \
            .mock(return_value=Response(500))
        try:
            await domain_client.request_api(opts)
            pytest.fail()
        except Exception as err:
            assert err.__class__.__name__ == 'InternalException'

    @respx.mock
    @pytest.mark.asyncio
    async def test_execute_normal_request(self):
        """Should execute request."""
        opts = {
            'url':  'https://mt-provisioning-api-v1.agiliumtrade.agiliumtrade.ai/users/current/accounts/accountId',
            'method': 'GET',
            'headers': {
                'auth-token': token
            },
        }

        request_stub = respx.get('https://mt-provisioning-api-v1.agiliumtrade.agiliumtrade.ai/users/current/'
                                 'accounts/accountId') \
            .mock(return_value=Response(200, content=json.dumps(expected)))
        response = await domain_client.request(opts)
        assert response == expected
        assert request_stub.call_count == 1



from .drawdownListener import DrawdownListener
from .drawdownListenerManager import DrawdownListenerManager
from ..domain_client import DomainClient
from mock import MagicMock, patch, AsyncMock
from asyncio import sleep
import pytest

token = 'header.payload.sign'
expected = [{
    'sequenceNumber': 2,
    'accountId': 'accountId',
    'trackerId': 'trackerId',
    'period': 'day',
    'startBrokerTime': '2022-04-08 00:00:00.000',
    'endBrokerTime': '2022-04-08 23:59:59.999',
    'brokerTime': '2022-04-08 09:36:00.000',
    'absoluteDrawdown': 250,
    'relativeDrawdown': 0.25
}, {
    'sequenceNumber': 3,
    'accountId': 'accountId',
    'trackerId': 'trackerId',
    'period': 'day',
    'startBrokerTime': '2022-04-08 00:00:00.000',
    'endBrokerTime': '2022-04-08 23:59:59.999',
    'brokerTime': '2022-04-08 09:36:00.000',
    'absoluteDrawdown': 250,
    'relativeDrawdown': 0.25
}]

expected2 = [{
    'sequenceNumber': 4,
    'accountId': 'accountId',
    'trackerId': 'trackerId',
    'period': 'day',
    'startBrokerTime': '2022-04-08 00:00:00.000',
    'endBrokerTime': '2022-04-08 23:59:59.999',
    'brokerTime': '2022-04-08 09:36:00.000',
    'absoluteDrawdown': 250,
    'relativeDrawdown': 0.25
}, {
    'sequenceNumber': 5,
    'accountId': 'accountId',
    'trackerId': 'trackerId',
    'period': 'day',
    'startBrokerTime': '2022-04-08 00:00:00.000',
    'endBrokerTime': '2022-04-08 23:59:59.999',
    'brokerTime': '2022-04-08 09:36:00.000',
    'absoluteDrawdown': 250,
    'relativeDrawdown': 0.25
}]
domain_client = DomainClient(MagicMock(), token, 'risk-management-api-v1')
drawdown_listener_manager = DrawdownListenerManager(domain_client)
call_stub = MagicMock()
listener = DrawdownListener()


@pytest.fixture(autouse=True)
async def run_around_tests():
    global domain_client
    domain_client = DomainClient(MagicMock(), token, 'risk-management-api-v1')
    global drawdown_listener_manager
    drawdown_listener_manager = DrawdownListenerManager(domain_client)
    global call_stub
    call_stub = MagicMock()

    class Listener(DrawdownListener):
        async def on_drawdown(self, drawdown_event):
            call_stub(drawdown_event)

    global listener
    listener = Listener()

    async def get_drawdown_func(arg, arg2):
        if arg == {
            'url': '/users/current/drawdown-events/stream',
            'method': 'GET',
            'qs': {
                'previousSequenceNumber': 1,
                'accountId': 'accountId',
                'trackerId': 'trackerId',
                'limit': 1000
            }
        }:
            await sleep(0.1)
            return expected
        elif arg == {
            'url': '/users/current/drawdown-events/stream',
            'method': 'GET',
            'qs': {
                'previousSequenceNumber': 3,
                'accountId': 'accountId',
                'trackerId': 'trackerId',
                'limit': 1000
            }
        }:
            await sleep(0.1)
            return expected2
        else:
            await sleep(0.1)
            return []

    get_drawdown_mock = AsyncMock(side_effect=get_drawdown_func)
    domain_client.request_api = get_drawdown_mock


class TestTradingClient:

    @pytest.mark.asyncio
    async def test_add_drawdown_listener(self):
        """Should add drawdown listener."""
        with patch('lib.clients.equityTracking.drawdownListenerManager.asyncio.sleep', new=lambda x: sleep(x / 10)):
            id = drawdown_listener_manager.add_drawdown_listener(listener, 'accountId', 'trackerId', 1)
            await sleep(0.22)
            assert call_stub.call_args_list[0][0][0] == expected[0]
            assert call_stub.call_args_list[1][0][0] == expected[1]
            assert call_stub.call_args_list[2][0][0] == expected2[0]
            assert call_stub.call_args_list[3][0][0] == expected2[1]
            drawdown_listener_manager.remove_drawdown_listener(id)

    @pytest.mark.asyncio
    async def test_remove_drawdown_listener(self):
        """Should remove drawdown listener."""
        with patch('lib.clients.equityTracking.drawdownListenerManager.asyncio.sleep', new=lambda x: sleep(x / 10)):
            id = drawdown_listener_manager.add_drawdown_listener(listener, 'accountId', 'trackerId', 1)
            await sleep(0.08)
            drawdown_listener_manager.remove_drawdown_listener(id)
            await sleep(0.22)
            assert call_stub.call_args_list[0][0][0] == expected[0]
            assert call_stub.call_args_list[1][0][0] == expected[1]
            assert call_stub.call_count == 2

    @pytest.mark.asyncio
    async def test_wait_if_error_returned(self):
        """Should wait if error returned."""
        call_count = 0

        async def get_drawdown_func(arg, arg2):
            nonlocal call_count
            call_count += 1
            if call_count < 3:
                raise Exception('test')

            if arg == {
                'url': '/users/current/drawdown-events/stream',
                'method': 'GET',
                'qs': {
                    'previousSequenceNumber': 1,
                    'accountId': 'accountId',
                    'trackerId': 'ABCD',
                    'limit': 1000
                }
            }:
                await sleep(0.05)
                return expected
            else:
                await sleep(0.5)
                return []

        get_drawdown_mock = AsyncMock(side_effect=get_drawdown_func)
        domain_client.request_api = get_drawdown_mock
        with patch('lib.clients.equityTracking.drawdownListenerManager.asyncio.sleep', new=lambda x: sleep(x / 10)):
            id = drawdown_listener_manager.add_drawdown_listener(listener, 'accountId', 'ABCD', 1)
            await sleep(0.06)
            assert domain_client.request_api.call_count == 1
            assert call_stub.call_count == 0
            await sleep(0.06)
            assert domain_client.request_api.call_count == 2
            assert call_stub.call_count == 0
            await sleep(0.2)
            assert domain_client.request_api.call_count == 3
            assert call_stub.call_count == 0
            await sleep(0.08)
            assert call_stub.call_count == 2
            call_stub.assert_any_call(expected[0])
            call_stub.assert_any_call(expected[1])
            drawdown_listener_manager.remove_drawdown_listener(id)



from .errorHandler import UnauthorizedException, ForbiddenException, ApiException, ConflictException, \
    ValidationException, InternalException, NotFoundException, TooManyRequestsException
from typing_extensions import TypedDict
from typing import Optional
from ..models import ExceptionMessage, date
from .timeoutException import TimeoutException
import json
import asyncio
import sys
import httpx
from datetime import datetime
from httpx import HTTPError, Response


if sys.version_info[0] == 3 and sys.version_info[1] >= 8 and sys.platform.startswith('win'):
    asyncio.set_event_loop_policy(asyncio.WindowsSelectorEventLoopPolicy())


class RequestOptions(TypedDict):
    """Options for HttpClient requests."""
    method: Optional[str]
    url: str
    headers: Optional[dict]
    params: Optional[dict]
    body: Optional[dict]
    files: Optional[dict]
    timeout: Optional[float]


class HttpClient:
    """HTTP client library based on requests module."""
    def __init__(self, timeout: float = 10, extended_timeout: float = 70, retry_opts=None):
        """Inits HttpClient class instance.

        Args:
            timeout: Request timeout in seconds.
            extended_timeout: Extended request timeout in seconds.
            retry_opts: Retry options.
        """
        if retry_opts is None:
            retry_opts = {}
        self._timeout = timeout
        self._extendedTimeout = extended_timeout
        self._retries = retry_opts['retries'] if 'retries' in retry_opts else 5
        self._minRetryDelayInSeconds = retry_opts['minDelayInSeconds'] if 'minDelayInSeconds' in retry_opts else 1
        self._maxRetryDelayInSeconds = retry_opts['maxDelayInSeconds'] if 'maxDelayInSeconds' in retry_opts else 30

    async def request(self, options: dict, is_extended_timeout: bool = False):
        """Performs a request. Response errors are returned as ApiError or subclasses.

        Args:
            options: Request options.
            is_extended_timeout: Whether to run the request with an extended timeout.

        Returns:
            Request result.
        """
        options['timeout'] = self._extendedTimeout if is_extended_timeout else self._timeout
        try:
            response = await self._make_request(options)
            response.raise_for_status()
            if response.content:
                try:
                    response = response.json()
                except Exception as err:
                    print('Error parsing json', err)
        except HTTPError as err:
            raise self._convert_error(err)
        return response

    async def request_with_failover(self, options: RequestOptions, retry_counter: int = 0, end_time: float = None) \
            -> Response:
        """Performs a request. Response errors are returned as ApiException or subclasses.

        Args:
            options: Request options.

        Returns:
            A request response.
        """
        options['timeout'] = self._timeout
        if not end_time:
            end_time = datetime.now().timestamp() + self._maxRetryDelayInSeconds * self._retries
        retry_after_seconds = 0
        try:
            response = await self._make_request(options)
            response.raise_for_status()
            if response.status_code == 202:
                retry_after_seconds = response.headers['retry-after']
                if isinstance(retry_after_seconds, str):
                    retry_after_seconds = float(retry_after_seconds)
            if response.content:
                try:
                    response = response.json()
                except Exception as err:
                    print('Error parsing json', err)
        except HTTPError as err:
            retry_counter = await self._handle_error(err, retry_counter, end_time)
            return await self.request_with_failover(options, retry_counter, end_time)
        if retry_after_seconds:
            await self._handle_retry(end_time, retry_after_seconds)
            response = await self.request_with_failover(options, retry_counter, end_time)
        return response

    async def _make_request(self, options: RequestOptions) -> Response:
        timeout = options['timeout'] if 'timeout' in options else self._timeout
        async with httpx.AsyncClient(timeout=timeout) as client:
            method = options['method'] if ('method' in options) else 'GET'
            url = options['url']
            params = options['params'] if 'params' in options else None
            files = options['files'] if 'files' in options else None
            headers = options['headers'] if 'headers' in options else None
            body = options['body'] if 'body' in options else None
            req = client.build_request(method, url, params=params, files=files, headers=headers, json=body)
            response = await client.send(req)
            return response

    async def _handle_retry(self, end_time: float, retry_after: float):
        if end_time > datetime.now().timestamp() + retry_after:
            await asyncio.sleep(retry_after)
        else:
            raise TimeoutException('Timed out waiting for the response')

    async def _handle_error(self, err, retry_counter: int, end_time: float):
        if err.__class__.__name__ == 'ConnectTimeout':
            error = err
        else:
            error = self._convert_error(err)
        if error.__class__.__name__ in ['ConflictException', 'InternalException', 'ApiException', 'ConnectTimeout'] \
                and retry_counter < self._retries:
            pause = min(pow(2, retry_counter) * self._minRetryDelayInSeconds, self._maxRetryDelayInSeconds)
            await asyncio.sleep(pause)
            return retry_counter + 1
        elif error.__class__.__name__ == 'TooManyRequestsException':
            retry_time = date(error.metadata['recommendedRetryTime']).timestamp()
            if retry_time < end_time:
                await asyncio.sleep(retry_time - datetime.now().timestamp())
                return retry_counter
        raise error

    def _convert_error(self, err: HTTPError):
        if err.__class__.__name__ == 'ConnectTimeout':
            return err
        try:
            response: ExceptionMessage or TypedDict = json.loads(err.response.text)
        except Exception:
            response = {}
        err_message = response['message'] if 'message' in response else (
            err.response.reason_phrase if hasattr(err, 'response') else None)
        status = None
        if hasattr(err, 'response'):
            status = err.response.status_code
        if status == 400:
            details = response['details'] if 'details' in response else []
            return ValidationException(err_message, details)
        elif status == 401:
            return UnauthorizedException(err_message)
        elif status == 403:
            return ForbiddenException(err_message)
        elif status == 404:
            return NotFoundException(err_message)
        elif status == 409:
            return ConflictException(err_message)
        elif status == 429:
            err_metadata = response['metadata'] if 'metadata' in response else {}
            return TooManyRequestsException(err_message, err_metadata)
        elif status == 500:
            return InternalException(err_message)
        else:
            return ApiException(err_message, status)



class TimeoutException(Exception):
    """Exception which indicates a timeout."""

    def __init__(self, message: str):
        """Inits the timeout exception

        Args:
            message: Exception message.
        """
        super().__init__(message)



from ..domain_client import DomainClient
from ...models import random_id
from .drawdownListener import DrawdownListener
import asyncio


class DrawdownListenerManager:
    """Drawdown event listener manager."""

    def __init__(self, domain_client: DomainClient):
        """Inits drawdown listener manager instance.

        Args:
            domain_client: Domain client.
        """
        self._domainClient = domain_client
        self._drawdownListeners = {}
        self._errorThrottleTime = 1

    @property
    def drawdown_listeners(self):
        """Returns the dictionary of drawdown listeners.

        Returns:
            Dictionary of drawdown listeners.
        """
        return self._drawdownListeners

    def add_drawdown_listener(self, listener: DrawdownListener, account_id: str = None, tracker_id: str = None,
                              sequence_number: int = None) -> str:
        """Adds a drawdown listener.

        Args:
            listener: Drawdown listener.
            account_id: Account id.
            tracker_id: Tracker id.
            sequence_number: Event sequence number.

        Returns:
            Drawdown listener id.
        """
        listener_id = random_id(10)
        self._drawdownListeners[listener_id] = listener
        asyncio.create_task(self._start_drawdown_event_job(listener_id, listener, account_id, tracker_id,
                                                           sequence_number))
        return listener_id

    def remove_drawdown_listener(self, listener_id: str):
        """Removes drawdown listener by id.

        Args:
            listener_id: Listener id.
        """
        if listener_id in self._drawdownListeners:
            del self._drawdownListeners[listener_id]

    async def _start_drawdown_event_job(self, listener_id: str, listener: DrawdownListener, account_id: str = None,
                                        tracker_id: str = None, sequence_number: int = None):
        throttle_time = self._errorThrottleTime
        while listener_id in self._drawdownListeners:
            opts = {
                'url': '/users/current/drawdown-events/stream',
                'method': 'GET',
                'qs': {
                    'previousSequenceNumber': sequence_number,
                    'accountId': account_id,
                    'trackerId': tracker_id,
                    'limit': 1000
                },
            }
            try:
                packets = await self._domainClient.request_api(opts, True)
                for packet in packets:
                    await listener.on_drawdown(packet)
                throttle_time = self._errorThrottleTime
                if listener_id in self._drawdownListeners and len(packets):
                    sequence_number = packets[-1]['sequenceNumber']
            except Exception as err:
                await asyncio.sleep(throttle_time)
                throttle_time = min(throttle_time * 2, 30)



from abc import abstractmethod


class DrawdownListener:
    """Drawdown listener for handling a stream of drawdown events."""

    @abstractmethod
    async def on_drawdown(self, drawdown_event):
        """Processes drawdown event which occurs when a drawdown limit is exceeded in a drawdown tracker.

        Args:
            drawdown_event: Drawdown event.
        """
        pass



from datetime import datetime
from typing_extensions import TypedDict
from typing import List, Optional
import iso8601
import random
import string
import pytz
import traceback


def date(date_time: str or float or int or datetime) -> datetime:
    """Parses a date string into a datetime object."""
    if isinstance(date_time, float) or isinstance(date_time, int):
        return datetime.fromtimestamp(max(date_time, 100000)).astimezone(pytz.utc)
    elif isinstance(date_time, datetime):
        return date_time
    else:
        return iso8601.parse_date(date_time)


def format_date(date_object: datetime or str) -> str:
    """Converts date to format compatible with JS"""
    if isinstance(date_object, datetime):
        return date_object.astimezone(pytz.utc).isoformat(timespec='milliseconds').replace('+00:00', 'Z')
    else:
        return date_object


def random_id(length: int = 32) -> str:
    """Generates a random id of 32 symbols."""
    return ''.join(random.choice(string.ascii_lowercase) for i in range(length))


class ValidationDetails(TypedDict):
    """Object to supply additional information for validation exceptions."""
    parameter: str
    """Name of invalid parameter."""
    value: Optional[str]
    """Entered invalid value."""
    message: str
    """Error message."""


class ExceptionMessage(TypedDict):
    """A REST API response that contains an exception message"""
    id: int
    """Error id"""
    error: str
    """Error name"""
    numericCode: Optional[int]
    """Numeric error code"""
    stringCode: Optional[str]
    """String error code"""
    message: str
    """Human-readable error message"""
    details: Optional[List[ValidationDetails]]
    """Additional information about error. Used to supply validation error details."""


def convert_iso_time_to_date(data):
    """Converts time fields of incoming data into datetime."""
    if not isinstance(data, str):
        for field in data:
            if isinstance(data,  dict):
                value = data[field]
            else:
                value = field
            if isinstance(value, str) and field in ['closeAfter', 'stoppedAt', 'stoppedTill', 'startTime', 'time',
                                                    'updateTime']:
                data[field] = date(value)
            if isinstance(value, list):
                for item in value:
                    convert_iso_time_to_date(item)
            if isinstance(value, dict):
                convert_iso_time_to_date(value)


def format_request(data: dict or list):
    """Formats datetime fields of a request into iso format."""
    if not isinstance(data, str):
        for field in data:
            if isinstance(data, dict):
                value = data[field]
            else:
                value = field
            if isinstance(value, datetime):
                data[field] = format_date(value)
            elif isinstance(value, list):
                for item in value:
                    format_request(item)
            elif isinstance(value, dict):
                format_request(value)


def format_error(err: Exception or any):
    """Formats and outputs metaApi errors with additional information.

    Args:
        err: Exception to process.
    """
    error = {'name': err.__class__.__name__, 'message': err if isinstance(err, str) or err is None else (
        err.args[0] if len(err.args) else None)}
    if hasattr(err, 'status_code'):
        error['status_code'] = err.status_code
    if err.__class__.__name__ == 'ValidationException':
        error['details'] = err.details
    if err.__class__.__name__ == 'TradeException':
        error['string_code'] = err.stringCode
    if err.__class__.__name__ == 'TooManyRequestsException':
        error['metadata'] = err.metadata
    error['trace'] = traceback.format_exc()
    return error



import setuptools

with open("README.rst", "r") as fh:
    long_description = fh.read()

install_requires = [
   'aiohttp==3.7.4', 'typing-extensions~=3.10.0.0', 'iso8601', 'pytz', 'requests==2.24.0', 'httpx==0.16.1'
]

tests_require = [
      'pytest', 'pytest-mock', 'pytest-asyncio', 'asynctest', 'aiohttp', 'mock', 'freezegun==1.0.0', 'respx==0.16.3'
]

setuptools.setup(
    name="metaapi_cloud_risk_management_sdk",
    version="1.1.1",
    author="Agilium Labs LLC",
    author_email="agiliumtrade@agiliumtrade.ai",
    description="Python SDK for MetaApi risk management API. Can execute trading risk restrictions, forex challenges "
                "and competitions in a cloud on both MetaTrader 5 (MT5) and MetaTrader 4 (MT4) (https://metaapi.cloud)",
    long_description=long_description,
    long_description_content_type="text/x-rst",
    keywords=['metaapi.cloud', 'MetaTrader', 'MetaTrader 5', 'MetaTrader 4', 'MetaTrader5', 'MetaTrader4', 'MT', 'MT4',
              'MT5', 'forex', 'equity tracking', 'risk management', 'API', 'REST', 'client', 'sdk', 'cloud', 'ftmo',
              'prop trading', 'proprietary trading'],
    url="https://github.com/agiliumtrade-ai/metaapi-risk-management-python-sdk",
    include_package_data=True,
    package_dir={'metaapi_cloud_risk_management_sdk': 'lib'},
    packages=['metaapi_cloud_risk_management_sdk'],
    install_requires=install_requires,
    tests_require=tests_require,
    license='SEE LICENSE IN LICENSE',
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: Other/Proprietary License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
)


from .drawdownListenerManager import DrawdownListenerManager
from .drawdownListener import DrawdownListener
from typing import List, TypedDict, Optional
from ..domain_client import DomainClient
from enum import Enum


class DrawdownTrackerUpdate(TypedDict, total=False):
    """Drawdown tracker configuration update."""
    name: str
    """Drawdown tracker name."""


class Period(Enum):
    """Period length to track drawdown for."""
    day = 'day'
    date = 'date'
    week = 'week'
    week_to_date = 'week-to-date'
    month = 'month'
    month_to_date = 'month-to-date'
    quarter = 'quarter'
    quarter_to_date = 'quarter-to-date'
    year = 'year'
    year_to_date = 'year-to-date'
    lifetime = 'lifetime'


class NewDrawdownTracker(DrawdownTrackerUpdate, total=False):
    """New drawdown tracker configuration."""
    startBrokerTime: Optional[str]
    """Time to start tracking from in broker timezone, YYYY-MM-DD HH:mm:ss.SSS format."""
    endBrokerTime: Optional[str]
    """Time to end tracking at in broker timezone, YYYY-MM-DD HH:mm:ss.SSS format."""
    period: Period
    """Period length to track drawdown for."""
    relativeDrawdownThreshold: Optional[float]
    """Relative drawdown threshold after which drawdown event is generated, a fraction of 1."""
    absoluteDrawdownThreshold: Optional[float]
    """Absolute drawdown threshold after which drawdown event is generated, should be greater than 0."""


class DrawdownTrackerId(TypedDict):
    """Drawdown tracker id."""
    id: str
    """Drawdown tracker id."""


class DrawdownTracker(NewDrawdownTracker):
    """Drawdown tracker configuration."""
    _id: str
    """Unique drawdown tracker id."""


class DrawdownEvent(TypedDict, total=False):
    """Drawdown threshold exceeded event model."""
    sequenceNumber: int
    """Event unique sequence number."""
    accountId: str
    """MetaApi account id."""
    trackerId: str
    """Drawdown tracker id."""
    startBrokerTime: str
    """Drawdown tracking period start time in broker timezone, in YYYY-MM-DD HH:mm:ss.SSS format."""
    endBrokerTime: Optional[str]
    """Drawdown tracking period end time in broker timezone, in YYYY-MM-DD HH:mm:ss.SSS format."""
    period: Period
    """Drawdown tracking period."""
    brokerTime: str
    """Drawdown threshold exceeded event time in broker timezone, in YYY-MM-DD HH:mm:ss.SSS format."""
    absoluteDrawdown: float
    """Absolute drawdown value which was observed when the drawdown threshold was exceeded."""
    relativeDrawdown: float
    """Relative drawdown value which was observed when the drawdown threshold was exceeded."""


class DrawdownPeriodStatistics(TypedDict, total=False):
    """Drawdown period statistics."""
    startBrokerTime: str
    """Period start time in broker timezone, in YYYY-MM-DD HH:mm:ss format."""
    endBrokerTime: Optional[str]
    """Period end time in broker timezone, in YYYY-MM-DD HH:mm:ss format."""
    period: Period
    """Period length."""
    initialBalance: float
    """Balance at period start time."""
    maxDrawdownTime: Optional[str]
    """Time max drawdown was observed at in broker timezone, in YYYY-MM-DD HH:mm:ss format"""
    maxAbsoluteDrawdown: Optional[float]
    """The value of maximum absolute drawdown observed."""
    maxRelativeDrawdown: Optional[float]
    """The value of maximum relative drawdown observed."""
    thresholdExceeded: bool
    """The flag indicating that max allowed total drawdown was exceeded."""


class EquityChartItem(TypedDict):
    """Equity chart item."""
    startBrokerTime: str
    """Start time of a chart item as per broker timezone, in YYYY-MM-DD HH:mm:ss format."""
    endBrokerTime: str
    """End time of a chart item as per broker timezone, in YYYY-MM-DD HH:mm:ss format."""
    averageBalance: float
    """Average balance value during the period."""
    minBalance: float
    """Minimum balance value during the period."""
    maxBalance: float
    """Maximum balance value during the period."""
    averageEquity: float
    """Average equity value during the period."""
    minEquity: float
    """Minimum equity value during the period."""
    maxEquity: float
    """Maximum equity value during the period."""


class EquityTrackingClient:
    """metaapi.cloud RiskManagement equity tracking API client (see https://metaapi.cloud/docs/riskManagement/)"""

    def __init__(self, domain_client: DomainClient):
        """Inits RiskManagement equity tracking API client instance.

        Args:
            domain_client: Domain client.
        """
        self._domainClient = domain_client
        self._drawdownListenerManager = DrawdownListenerManager(domain_client)

    async def create_drawdown_tracker(self, account_id: str, tracker: NewDrawdownTracker) -> 'DrawdownTrackerId':
        """Returns list of transactions on the strategies the current user provides to other users. See
        https://metaapi.cloud/docs/copyfactory/restApi/api/history/getProvidedTransactions/

        Args:
            account_id: Id of the MetaApi account.
            tracker: Drawdown tracker.

        Returns:
            A coroutine resolving with transactions found.
        """
        return await self._domainClient.request_api({
          'url': f'/users/current/accounts/{account_id}/drawdown-trackers',
          'method': 'POST',
          'body': tracker
        })

    async def get_drawdown_trackers(self, account_id: str) -> 'List[DrawdownTracker]':
        """Returns drawdown trackers defined for an account.

        Args:
            account_id: Id of the MetaApi account.

        Returns:
            A coroutine resolving with drawdown trackers.
        """
        return await self._domainClient.request_api({
            'url': f'/users/current/accounts/{account_id}/drawdown-trackers',
            'method': 'GET',
        })

    async def update_drawdown_tracker(self, account_id: str, id: str, update: DrawdownTrackerUpdate):
        """Updates drawdown tracker.

        Args:
            account_id: Id of the MetaApi account.
            id: Id of the drawdown tracker.
            update: Drawdown tracker update.

        Returns:
            A coroutine resolving when drawdown tracker updated.
        """
        return await self._domainClient.request_api({
            'url': f'/users/current/accounts/{account_id}/drawdown-trackers/{id}',
            'method': 'PUT',
            'body': update
        })

    async def delete_drawdown_tracker(self, account_id: str, id: str):
        """Removes drawdown tracker.

        Args:
            account_id: Id of the MetaApi account.
            id: Id of the drawdown tracker.

        Returns:
            A coroutine resolving when drawdown tracker is removed.
        """
        return await self._domainClient.request_api({
            'url': f'/users/current/accounts/{account_id}/drawdown-trackers/{id}',
            'method': 'DELETE'
        })

    async def get_drawdown_events(self, start_broker_time: str = None, end_broker_time: str = None,
                                  account_id: str = None, tracker_id: str = None, limit: int = None) \
            -> 'List[DrawdownEvent]':
        """Returns drawdown events by broker time range.

        Args:
            start_broker_time: Value of the event time in broker timezone to start loading data from, inclusive,
            in 'YYYY-MM-DD HH:mm:ss.SSS format.
            end_broker_time: Value of the event time in broker timezone to end loading data at, inclusive,
            in 'YYYY-MM-DD HH:mm:ss.SSS format.
            account_id: Id of the MetaApi account.
            tracker_id: Id of the drawdown tracker.
            limit: Pagination limit, default is 1000.

        Returns:
            A coroutine resolving with drawdown events.
        """
        qs = {}
        if start_broker_time is not None:
            qs['startBrokerTime'] = start_broker_time
        if end_broker_time is not None:
            qs['endBrokerTime'] = end_broker_time
        if account_id is not None:
            qs['accountId'] = account_id
        if tracker_id is not None:
            qs['trackerId'] = tracker_id
        if limit is not None:
            qs['limit'] = limit
        return await self._domainClient.request_api({
            'url': '/users/current/drawdown-events/by-broker-time',
            'method': 'GET',
            'params': qs
        })

    def add_drawdown_listener(self, listener: DrawdownListener, account_id: str = None, tracker_id: str = None,
                              sequence_number: int = None) -> str:
        """Adds a drawdown listener and creates a job to make requests.

        Args:
            listener: Drawdown listener.
            account_id: Account id.
            tracker_id: Tracker id.
            sequence_number: Sequence number.

        Returns:
            Listener id.
        """
        return self._drawdownListenerManager.add_drawdown_listener(listener, account_id, tracker_id, sequence_number)

    def remove_drawdown_listener(self, listener_id: str):
        """Removes drawdown listener and cancels the event stream

        Args:
            listener_id: Drawdown listener id.
        """
        self._drawdownListenerManager.remove_drawdown_listener(listener_id)

    async def get_drawdown_statistics(self, account_id: str, tracker_id: str, start_time: str = None,
                                      limit: int = None) -> 'List[DrawdownPeriodStatistics]':
        """Returns account drawdown tracking stats by drawdown tracker id.

        Args:
            account_id: Id of MetaAPI account.
            tracker_id: Id of drawdown tracker.
            start_time: Time to start loading stats from, default is current time. Note that stats is loaded in
            backwards direction.
            limit: Number of records to load, default is 1.

        Returns:
            A coroutine resolving with drawdown statistics.
        """
        qs = {}
        if start_time is not None:
            qs['startTime'] = start_time
        if limit is not None:
            qs['limit'] = limit
        return await self._domainClient.request_api({
            'url': f'/users/current/accounts/{account_id}/drawdown-trackers/{tracker_id}/statistics',
            'method': 'GET',
            'params': qs
        })

    async def get_equity_chart(self, account_id: str, start_time: str = None, end_time: str = None) \
            -> 'List[EquityChartItem]':
        """Returns equity chart by account id.

        Args:
            account_id: MetaApi account id.
            start_time: Starting broker time in YYYY-MM-DD HH:mm:ss format.
            end_time: Ending broker time in YYYY-MM-DD HH:mm:ss format.

        Returns:
            A coroutine resolving with equity chart.
        """
        qs = {}
        if start_time is not None:
            qs['startTime'] = start_time
        if end_time is not None:
            qs['endTime'] = end_time
        return await self._domainClient.request_api({
            'url': f'/users/current/accounts/{account_id}/equity-chart',
            'method': 'GET',
            'params': qs
        })



from metaapi_cloud_sdk import RiskManagement, DrawdownListener
import os
import asyncio
import json

# your MetaApi API token
token = os.getenv('TOKEN') or '<put in your token here>'
# your MetaApi account id
# the account must have field riskManagementApiEnabled set to true
account_id = os.getenv('ACCOUNT_ID') or '<put in your account id here>'
domain = os.getenv('DOMAIN')


async def main():
    risk_management = RiskManagement(token, {'domain': domain})
    risk_management_api = risk_management.risk_management_api

    class ExampleDrawdownListener(DrawdownListener):
        async def on_drawdown(self, drawdown_event):
            print('drawdown event received', drawdown_event)

    try:
        # creating a tracker
        tracker_id = await risk_management_api.create_drawdown_tracker(account_id, {
            'name': 'example-tracker',
            'absoluteDrawdownThreshold': 5,
            'period': 'day'
        })
        print('Created a drawdown tracker ' + tracker_id['id'])

        # adding a drawdown listener
        drawdown_listener = ExampleDrawdownListener()
        listener_id = risk_management_api.add_drawdown_listener(drawdown_listener, account_id, tracker_id['id'])

        print('Streaming drawdown events for 1 minute...')
        await asyncio.sleep(60)
        risk_management_api.remove_drawdown_listener(listener_id)

        print('Receiving statistics with REST API')
        events = await risk_management_api.get_drawdown_events(None, None, account_id, tracker_id['id'])
        print('drawdown events', json.dumps(events))
        statistics = await risk_management_api.get_drawdown_statistics(account_id, tracker_id['id'])
        print('drawdown statistics', json.dumps(statistics))
        equity_chart = await risk_management_api.get_equity_chart(account_id)
        print('equity chart', json.dumps(equity_chart))

        # removing the tracker
        await risk_management_api.delete_drawdown_tracker(account_id, tracker_id['id'])
        print('Removed the tracker')
    except Exception as err:
        print(risk_management.format_error(err))
    exit()

asyncio.run(main())
